from nonebot import on_command
from nonebot.plugin import PluginMetadata
from nonebot.params import Command, CommandArg
from nonebot.adapters.onebot.v11 import MessageEvent, Message, Bot, MessageSegment

from .data_source import make_midi, multi_tracks

__plugin_meta__ = PluginMetadata(
    name="在线编曲",
    description="在线编曲",
    usage="""
    usage：
    在线编曲
    仅支持简谱表达，升八度用+，降八度用-，延音用~，半音用_，附点.，可叠加，空格分隔
    升半音用#，降半音用b，休止符用0，不可叠加，空格分隔，顺序不影响
    乐器代号参照midi乐器列表，不是所有乐器都可用，0为大钢琴
    支持多音轨，用 `>` 表示新的音轨，轨道范围为0-15，力度为0-1的小数
    BPM一般为120，调号一般为C，小调加m，例子：C C# C#m Cb Cm
    指令：
        编曲 [乐器代号] [BPM] [调号] [简谱]
        编曲 [BPM] [调号] > [轨道] [乐器] [力度] [简谱] > [轨道] [乐器] [力度] [简谱]
        编曲 0 120 C 3 5 6 1+ 2+~ 1+ 2+ 3+~ 1+ 6 5 3 1+ 2+ 6~
        """.strip(),
    extra={
        "unique_name": "make_midi",
        "author": "RandomEnch <randomench@foxmail.com>",
    },
)

make_music = on_command("编曲", priority=5, block=True)


@make_music.handle()
async def _(event: MessageEvent, arg: Message = CommandArg()):
    if not arg:
        await make_music.finish("未输入参数")
    qq = str(event.user_id)
    arg = arg.extract_plain_text()
    if '>' not in arg:
        arg = arg.split()
        program = int(arg[0])
        bpm = int(arg[1])
        key_signature = arg[2]
        notes = arg[3:]
        try:
            result = make_midi(qq, notes, program=program, bpm=bpm, key_signature=key_signature)
        except Exception as e:
            result = f"编曲失败，参数错误：{e}"
    else:
        arg = arg.replace('\n', '').split('>')
        bpm = int(arg[0].split()[0])
        key_signature = arg[0].split()[1]
        tracks = arg[1:]
        try:
            result = multi_tracks(qq, tracks, bpm=bpm, key_signature=key_signature)
        except Exception as e:
            result = f"编曲失败，参数错误：{e}"

    await make_music.finish(result)





# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot_plugin_makemidi']

package_data = \
{'': ['*'], 'nonebot_plugin_makemidi': ['resources/*']}

install_requires = \
['fluidsynth>=0.2,<0.3',
 'midi2audio>=0.1.1,<0.2.0',
 'mido>=1.2.10,<2.0.0',
 'nonebot-adapter-onebot>=2.0.0-beta.1,<3.0.0',
 'nonebot2>=2.0.0-beta.4,<3.0.0',
 'pydub>=0.25.1,<0.26.0']

setup_kwargs = {
    'name': 'nonebot-plugin-makemidi',
    'version': '0.1.11',
    'description': 'Easy midi maker',
    'long_description': None,
    'author': 'RandomEnch',
    'author_email': 'randomench@foxmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
from pathlib import Path
from pydub import AudioSegment
from midi2audio import FluidSynth
from mido import Message, MidiFile, MidiTrack, MetaMessage, bpm2tempo
from nonebot.adapters.onebot.v11 import MessageSegment

current_path = Path(__file__).resolve().parent / "resources"
midi_path = current_path / 'midi'


def signature(key_signature, note, tone_change):
    if key_signature == 'C' or key_signature == 'Am':
        pass
    elif key_signature == 'G' or key_signature == 'Em':
        tones = [5, 6, 7, 1, 2, 3, 4]
        note = tones[note - 1]
        if note == 4:
            tone_change += 1
    elif key_signature == 'F' or key_signature == 'Dm':
        tones = [4, 5, 6, 7, 1, 2, 3]
        note = tones[note - 1]
        if note == 7:
            tone_change -= 1
    elif key_signature == 'D' or key_signature == 'Bm':
        tones = [2, 3, 4, 5, 6, 7, 1]
        note = tones[note - 1]
        if note == 1 or note == 4:
            tone_change += 1
    elif key_signature == 'Bb' or key_signature == 'Gm':
        tones = [7, 1, 2, 3, 4, 5, 6]
        note = tones[note - 1]
        if note == 7 or note == 3:
            tone_change -= 1
    elif key_signature == 'A' or key_signature == 'F#m':
        tones = [6, 7, 1, 2, 3, 4, 5]
        note = tones[note - 1]
        if note == 1 or note == 4 or note == 5:
            tone_change += 1
    elif key_signature == 'Eb' or key_signature == 'Cm':
        tones = [3, 4, 5, 6, 7, 1, 2]
        note = tones[note - 1]
        if note == 7 or note == 3 or note == 6:
            tone_change -= 1
    elif key_signature == 'E' or key_signature == 'C#m':
        tones = [3, 4, 5, 6, 7, 1, 2]
        note = tones[note - 1]
        if note == 1 or note == 4 or note == 5 or note == 2:
            tone_change += 1
    elif key_signature == 'Ab' or key_signature == 'Fm':
        tones = [6, 7, 1, 2, 3, 4, 5]
        note = tones[note - 1]
        if note == 7 or note == 3 or note == 6 or note == 2:
            tone_change -= 1
    elif key_signature == 'B' or key_signature == 'G#m' or key_signature == 'Cb':
        tones = [7, 1, 2, 3, 4, 5, 6]
        note = tones[note - 1]
        if note == 1 or note == 4 or note == 5 or note == 2 or note == 6:
            tone_change += 1
    elif key_signature == 'C#' or key_signature == 'Db' or key_signature == 'A#m' or key_signature == 'Bbm':
        tones = [1, 2, 3, 4, 5, 6, 7]
        note = tones[note - 1]
        tone_change += 1
    elif key_signature == 'F#' or key_signature == 'D#m' or key_signature == 'Gb':
        tones = [4, 5, 6, 7, 1, 2, 3]
        note = tones[note - 1]
        if note != 7:
            tone_change += 1
    else:
        pass
    return note, tone_change


def parser_notes(note, key_signature):
    if '~' in note:
        length = 1 + int(str(note).count('~'))
        note = note.replace('~', '')
    elif '_' in note:
        length = 1 * 0.5 ** int(str(note).count('_'))
        note = note.replace('_', '')
    elif '.' in note:
        length = 1 + 0.5 * int(str(note).count('.'))
        note = note.replace('.', '')
    else:
        length = 1
    if '#' in note:
        tone_change = 1
        note = note.replace('#', '')
    elif 'b' in note:
        tone_change = -1
        note = note.replace('b', '')
    else:
        tone_change = 0
    if '+' in note:
        base_sum = int(str(note).count('+'))
        note = int(note.replace('+', ''))
    elif '-' in note:
        base_sum = -int(str(note).count('-'))
        note = int(note.replace('-', ''))
    else:
        note = int(note)
        base_sum = 0
    # 小调
    if 'm' in key_signature:
        if note == 3 or note == 6 or note == 7:
            tone_change -= 1

    return note, length, tone_change, base_sum


def play_note(note, length, track, bpm=120, base_num=0, delay=0, velocity=1.0, channel=0, tone_change=0):
    meta_time = 60 * 60 * 10 / bpm
    major_notes = [0, 2, 2, 1, 2, 2, 2, 1]
    base_note = 60
    if note != 0 and 1 <= note <= 7:
        if tone_change == 0:
            note = base_note + base_num * 12 + sum(major_notes[0:note])
        elif tone_change > 0:
            note = base_note + base_num * 12 + sum(major_notes[0:note]) + tone_change
        elif tone_change < 0:
            note = base_note + base_num * 12 + sum(major_notes[0:note]) + tone_change
        track.append(
            Message('note_on', note=note, velocity=round(64 * velocity),
                    time=round(delay * meta_time), channel=channel))
        track.append(
            Message('note_off', note=note, velocity=round(64 * velocity),
                    time=round(meta_time * length), channel=channel))
    elif note == 0:
        track.append(
            Message('note_on', note=note, velocity=round(64 * velocity),
                    time=round(delay * meta_time), channel=channel))
        track.append(
            Message('note_off', note=note, velocity=round(64 * velocity),
                    time=round(meta_time * length), channel=channel))


def make_midi(qq, notes, bpm=120, program=0, key_signature='C'):
    if os.path.exists(midi_path):
        pass
    else:
        os.mkdir(midi_path)
    try:
        os.remove(midi_path / f'{qq}.mid')
        os.remove(midi_path / f'{qq}.wav')
    except:
        pass
    mid = MidiFile()
    track = MidiTrack()
    mid.tracks.append(track)
    tempo = bpm2tempo(bpm)
    track.append(Message('program_change', channel=0, program=program, time=0))
    track.append(MetaMessage('set_tempo', tempo=tempo, time=0))
    track.append(MetaMessage('key_signature', key=key_signature))
    for note in notes:
        note, length, tone_change, base_sum = parser_notes(note, key_signature)
        note, tone_change = signature(key_signature, note, tone_change)
        play_note(note, length, track, bpm, base_sum, tone_change=tone_change)

    mid.save(midi_path / f'{qq}.mid')
    midi2wav(qq)
    high_volume(qq)
    return MessageSegment.record(midi_path / f'{qq}.wav')


def multi_tracks(qq, tracks, bpm=120, key_signature='C'):
    if os.path.exists(midi_path):
        pass
    else:
        os.mkdir(midi_path)
    try:
        os.remove(midi_path / f'{qq}.mid')
        os.remove(midi_path / f'{qq}.wav')
    except:
        pass
    mid = MidiFile(type=1)
    tempo = bpm2tempo(bpm)
    for simple in tracks:
        if simple[0] == ' ':
            simple = simple[1:]
        channel = int(simple.split()[0])
        program = int(simple.split()[1])
        velocity = float(simple.split()[2])
        notes = simple.split()[3:]
        track = MidiTrack()
        mid.tracks.append(track)
        track.append(MetaMessage('set_tempo', tempo=tempo, time=0))
        track.append(MetaMessage('key_signature', key=key_signature))
        track.append(Message('program_change', channel=channel, program=program, time=0))
        for note in notes:
            note, length, tone_change, base_sum = parser_notes(note, key_signature)
            note, tone_change = signature(key_signature, note, tone_change)
            play_note(note, length, track, bpm, base_sum, tone_change=tone_change, channel=channel, velocity=velocity)

    mid.save(midi_path / f'{qq}.mid')
    midi2wav(qq)
    high_volume(qq)
    return MessageSegment.record(midi_path / f'{qq}.wav')


def midi2wav(qq):
    sf_path = current_path / 'gm.sf2'
    s = FluidSynth(sound_font=sf_path)
    s.midi_to_audio(midi_path / f'{qq}.mid', midi_path / f'{qq}.wav')


def high_volume(qq):
    song = AudioSegment.from_wav(midi_path / f'{qq}.wav')
    song = song + 20
    song.export(midi_path / f'{qq}.wav', format="wav")
